package com.softslate.commerce.businessobjects.core;

import java.util.HashMap;
import java.util.Map;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;

import com.softslate.commerce.businessobjects.core.BusinessObjectFactory;
import com.softslate.commerce.businessobjects.product.ProductProcessor;
import com.softslate.commerce.businessobjects.product.CategoryProcessor;
import com.softslate.commerce.businessobjects.core.CommandLineSupport;
import com.softslate.commerce.daos.core.DAOFactory;

/**
 * 
 * A class which may be run from the command line to sort products or
 * categories.
 * 
 * This example uses the <code>CommandLineSupport</code> class to initialize all
 * of the system objects such as the <code>BusinessObjectFactory</code> and the
 * <code>DAOFactory</code>.
 * 
 * @author dtobey
 * 
 */
public class CommandLineSorter extends CommandLineSupport {

	static Log log = LogFactory.getLog(CommandLineSorter.class);

	/**
	 * Main method, which gets invoked when executing this class from the
	 * command line.
	 * 
	 * @param args
	 */
	public static void main(String[] args) {

		// Log the start of the process
		if (log.isInfoEnabled())
			log.info("Starting sorter.");

		// This command requires 2 arguments and may have as many as 4. The
		// first argument may be
		// "product" or "category" depending on which entity is
		// being sorted. The second parameter is either "code" or "name",
		// depending on which field is used for the resorting. The third
		// parameter only pertains to category sorting and may be "ContentPages"
		// or "Categories", depending on if the sorting is restricted to just
		// content pages or categories. The fourth parameter also only pertains
		// to category sorting. It contains the categoryID of the parent
		// category that the sorting should be restricted to.
		if (args == null || args.length < 2 || args.length > 4) {
			if (log.isErrorEnabled())
				log
						.error("Wrong number of arguments. 2, 3, or 4 arguments needed.");
			System.out
					.println("Usage: ./commandLineSorter.sh product|category code|name ContentPages|Categories [categoryID]");
			return;
		}

		// Initialize the system objects
		Map systemObjects = initializeSystem("Sorter Admin User");
		if (systemObjects == null) {
			if (log.isErrorEnabled())
				log.error("Initialization failed ... exiting.");
			return;
		}

		// Unpack the system objects needed by this process
		BusinessObjectFactory bof = (BusinessObjectFactory) systemObjects
				.get("bof");
		// Properties appSettings = (Properties)
		// systemObjects.get("appSettings");
		// Properties appComponents = (Properties) systemObjects
		// .get("appComponents");
		DAOFactory daoFactory = (DAOFactory) systemObjects.get("daoFactory");
		// Settings settings = (Settings) systemObjects.get("settings");
		// User user = (User) systemObjects.get("user");

		try {
			Map results = null;

			// Call the processor to do the sorting
			if (args[0].equals("product")) {
				ProductProcessor processor =  bof
						.createObject(ProductProcessor.class);
				if (args[1].equals("code")) {
					results = processor.sortProductsByCode(null);
				} else if (args[1].equals("name")) {
					results = processor.sortProductsByName(null);
				}

			} else if (args[0].equals("category")) {
				Map parameters = new HashMap();
				String workWithCategories = "Categories";
				if (args[2] != null && args[2].equals("ContentPages")) {
					workWithCategories = "Content Pages";
				}
				parameters.put("workWithCategories", workWithCategories);
				String parentCategory = args[3];
				if (parentCategory != null && parentCategory.length() > 0
						&& !parentCategory.equals("0")) {
					parameters.put("parentCategory", parentCategory);
				}
				if (log.isInfoEnabled())
					log.info("Sorter paramters: " +  parameters);
				CategoryProcessor processor =  bof
						.createObject(CategoryProcessor.class);
				if (args[1].equals("code")) {
					results = processor.sortCategoriesByCode(parameters);
				} else if (args[1].equals("name")) {
					results = processor.sortCategoriesByName(parameters);
				}
			} else if (args[0].equals("manufacturer")) {
				System.out.println("Manufacturer sorting not supported.");
			}

			// Commit the database transaction and close the Hibernate session
			if (daoFactory != null) {
				if (log.isDebugEnabled())
					log
							.debug("Committing Hibernate transaction (if still open).");
				daoFactory.commitTransaction();
				if (log.isDebugEnabled())
					log.debug("Closing Hibernate session (if open).");
				daoFactory.closeSession();
			}

			// Output the results
			if (log.isInfoEnabled())
				log.info("Sorting results: " + results);

		} catch (Exception e) {
			if (log.isErrorEnabled())
				log.error("Sorting failed ... exiting.", e);
			return;
		}

		// Log the end of process
		if (log.isInfoEnabled())
			log.info("Ending sorter.");
	}
}
