package com.softslate.commerce.businessobjects.product;

import java.util.Collection;
import java.util.HashMap;
import java.util.Map;

import org.apache.commons.beanutils.BeanUtils;
import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;

import com.softslate.commerce.businessobjects.core.BaseBusinessProcessor;
import com.softslate.commerce.daos.product.AttributeDAO;
import com.softslate.commerce.daos.product.AttributeGatewayDAO;
import com.softslate.commerce.daos.product.ProductGatewayDAO;
import com.softslate.commerce.daos.product.SKUGatewayDAO;

/**
 * Processes business logic having to do with attributes. Created in the Struts
 * layer in response to requests for viewing, editing, adding, and deleting
 * attributes.
 * <p>
 * <code>BasicAttributeProcessor</code> is the default implementation of the
 * <code>AttributeProcessor</code> interface for the application.
 * 
 * @author David Tobey
 * 
 */
public class BasicAttributeProcessor extends BaseBusinessProcessor implements
		AttributeProcessor {

	static Log log = LogFactory.getLog(BasicAttributeProcessor.class);

	public Collection loadAllAttributes() throws Exception {
		if (log.isInfoEnabled() && getUser().getAdministrator() != null)
			log.info(getUser().getAdministrator().getUserName()
					+ ": Loading all attributes");
		AttributeGatewayDAO gatewayDAO =  getDaoFactory()
				.createDAO(AttributeGatewayDAO.class);
		return gatewayDAO.loadAllAttributes();
	}

	public Map loadAttributesAndCount(Map parameters) throws Exception {
		if (log.isInfoEnabled() && getUser().getAdministrator() != null)
			log.info(getUser().getAdministrator().getUserName()
					+ ": Loading attributes and count");
		AttributeGatewayDAO gatewayDAO =  getDaoFactory()
				.createDAO(AttributeGatewayDAO.class);
		return gatewayDAO.loadAttributesAndCount(parameters);
	}

	public Map addAttribute(Map parameters) throws Exception {
		if (log.isInfoEnabled() && getUser().getAdministrator() != null) {
			String code = (String) parameters.get("attributeCode");
			log.info(getUser().getAdministrator().getUserName()
					+ ": Adding attribute: " + code);
		}
		Map results = new HashMap();

		// Check that another attribute with the same code doesn't exist
		Attribute attribute =  getBusinessObjectFactory()
				.createObject(Attribute.class);
		BeanUtils.copyProperties(attribute, parameters);
		results = validateAttribute(attribute);
		if (results != null && results.get("resultCode") != null
				&& !results.get("resultCode").equals("0")) {
			return results;
		}

		// Insert the attribute
		AttributeDAO dao = (AttributeDAO) getDaoFactory().createDAO(
				"attributeDAOImplementer");
		dao.setAttribute(attribute);
		dao.insertAttribute(true);
		results.put("newRecord", dao.getAttribute());

		// Insert the association with the parent product, if present
		if (parameters.get("productID") != null
				&& !parameters.get("productID").equals("")) {

			ProductProcessor processor =  getBusinessObjectFactory()
					.createObject(ProductProcessor.class);
			Product product = processor.loadProductFromID(parameters);

			if (product != null) {
				ProductAttribute pa =  getBusinessObjectFactory()
						.createObject(ProductAttribute.class);
				AttributeGatewayDAO gatewayDAO =  getDaoFactory()
						.createDAO(AttributeGatewayDAO.class);
				pa.setAttribute(attribute);
				pa.setProduct(product);
				gatewayDAO.insertProductAttribute(pa);
			}
		}
		return results;
	}

	public Attribute loadAttributeFromID(Map parameters) throws Exception {
		if (log.isInfoEnabled() && getUser().getAdministrator() != null)
			log.info(getUser().getAdministrator().getUserName()
					+ ": Loading attribute from id");
		Attribute attribute =  getBusinessObjectFactory()
				.createObject(Attribute.class);
		AttributeDAO dao = (AttributeDAO) getDaoFactory().createDAO(
				"attributeDAOImplementer");
		BeanUtils.copyProperties(attribute, parameters);
		dao.setAttribute(attribute);
		dao.loadAttributeFromID();
		return dao.getAttribute();
	}

	public Attribute loadAttributeFromCode(Map parameters) throws Exception {
		if (log.isInfoEnabled() && getUser().getAdministrator() != null)
			log.info(getUser().getAdministrator().getUserName() + ": Loading attribute from id");
		Attribute attribute = getBusinessObjectFactory().createObject(Attribute.class);
		AttributeDAO dao = (AttributeDAO) getDaoFactory().createDAO("attributeDAOImplementer");
		BeanUtils.copyProperties(attribute, parameters);
		dao.setAttribute(attribute);
		dao.loadAttributeFromCode();
		return dao.getAttribute();
	}

	public Map editAttribute(Map parameters) throws Exception {
		if (log.isInfoEnabled() && getUser().getAdministrator() != null) {
			String code = (String) parameters.get("attributeCode");
			log.info(getUser().getAdministrator().getUserName()
					+ ": Editing attribute: " + code);
		}
		Map results = new HashMap();

		// Check that another attribute with the same code doesn't exist
		if (parameters.get("attributeID") != null) {
			Attribute attribute = getBusinessObjectFactory().createObject(Attribute.class);
			BeanUtils.copyProperties(attribute, parameters);
			results = validateAttribute(attribute);
			if (results != null && results.get("resultCode") != null && !results.get("resultCode").equals("0")) {
				return results;
			}
		}

		// Load the attribute
		Attribute attribute = null;
		if (parameters.get("attributeID") != null) {
			attribute = loadAttributeFromID(parameters);
		}
		if (attribute == null && parameters.get("attributeCode") != null) {
			attribute = loadAttributeFromCode(parameters);
		}

		// Update the attribute
		BeanUtils.copyProperties(attribute, parameters);
		AttributeDAO dao = (AttributeDAO) getDaoFactory().createDAO(
				"attributeDAOImplementer");
		dao.setAttribute(attribute);
		dao.updateAttribute(true);
		results.put("updatedRecord", dao.getAttribute());
		String keepOrdersInSync = (String) getSettings().getValue(
				"keepOrdersInSync");
		if (keepOrdersInSync != null && keepOrdersInSync.equals("1")) {
			ProductGatewayDAO gatewayDao =  getDaoFactory()
					.createDAO(ProductGatewayDAO.class);
			gatewayDao.updateProductAttributeOrders();
		}
		return results;
	}

	public Map editDeleteAttributes(Map parameters) throws Exception {
		if (log.isInfoEnabled() && getUser().getAdministrator() != null)
			log.info(getUser().getAdministrator().getUserName()
					+ ": Editing/deleting multiple attributes");
		AttributeGatewayDAO dao =  getDaoFactory()
				.createDAO(AttributeGatewayDAO.class);
		Map results = dao.editDeleteAttributes(parameters, getBusinessObjectFactory());
		String keepOrdersInSync = (String) getSettings().getValue(
				"keepOrdersInSync");
		if (keepOrdersInSync != null && keepOrdersInSync.equals("1")) {
			ProductGatewayDAO gatewayDao =  getDaoFactory()
					.createDAO(ProductGatewayDAO.class);
			gatewayDao.updateProductAttributeOrders();
		}
		SKUGatewayDAO skuDao = (SKUGatewayDAO) getDaoFactory().createDAO(
				"skuGatewayDAOImplementer");
		skuDao.deleteOrphanedSKUs(true);
		return results;
	}

	public void deleteAttribute(Map parameters) throws Exception {
		if (log.isInfoEnabled() && getUser().getAdministrator() != null) {
			String code = (String) parameters.get("attributeCode");
			log.info(getUser().getAdministrator().getUserName()
					+ ": Deleting attribute: " + code);
		}
		Attribute attribute =  getBusinessObjectFactory()
				.createObject(Attribute.class);
		BeanUtils.copyProperties(attribute, parameters);
		AttributeDAO dao = (AttributeDAO) getDaoFactory().createDAO(
				"attributeDAOImplementer");
		dao.setAttribute(attribute);
		dao.deleteAttribute(false);
		SKUGatewayDAO skuDao = (SKUGatewayDAO) getDaoFactory().createDAO(
				"skuGatewayDAOImplementer");
		skuDao.deleteOrphanedSKUs(true);
		return;
	}

	public Map validateAttribute(Attribute attribute) throws Exception {
		Map results = new HashMap();

		// Check for a duplicate code
		AttributeDAO dao = (AttributeDAO) getDaoFactory().createDAO(
				"attributeDAOImplementer");
		dao.setAttribute(attribute);
		dao.loadAttributeFromCode();
		if (dao.getAttribute() != null) {
			dao.getDaoFactory().getSession().evict(dao.getAttribute());
		}
		boolean insertingButExists = attribute.getAttributeID() == 0
				&& dao.getAttribute() != null
				&& dao.getAttribute().getAttributeID() > 0;
		boolean updatingButExists = attribute.getAttributeID() > 0
				&& dao.getAttribute() != null
				&& attribute.getAttributeID() != dao.getAttribute()
						.getAttributeID();
		if (insertingButExists || updatingButExists) {
			results.put("resultCode", "DuplicateRecord");
			results.put("duplicateField", "attributeCode");
			results.put("existingRecord", dao.getAttribute());
		}
		return results;
	}

	public int loadMaxAttributeID() throws Exception {
		if (log.isInfoEnabled() && getUser().getAdministrator() != null) {
			log.info(getUser().getAdministrator().getUserName()
					+ ": Loading max attribute ID.");
		}
		AttributeGatewayDAO gatewayDAO =  getDaoFactory()
				.createDAO(AttributeGatewayDAO.class);
		return gatewayDAO.loadMaxAttributeID();
	}
}
