package com.softslate.commerce.customer.core;

import java.io.IOException;

import javax.servlet.Filter;
import javax.servlet.FilterChain;
import javax.servlet.FilterConfig;
import javax.servlet.RequestDispatcher;
import javax.servlet.ServletException;
import javax.servlet.ServletRequest;
import javax.servlet.ServletResponse;
import javax.servlet.http.HttpServletRequest;
import javax.servlet.http.HttpServletResponse;

import net.sf.navigator.menu.RolesPermissionsAdapter;

import org.apache.commons.lang.WordUtils;
import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;

import com.softslate.commerce.administrator.core.AdministratorRequestWrapper;
import com.softslate.commerce.businessobjects.core.Settings;

/**
 * Maps /api requests to the appropriate .do URI. This can be thought of as the first half of a wrapper surrounding
 * Struts for API calls. This part of the wrapper maps request under /api to the .do URLs that Struts needs.
 * 
 * The second half of the wrapper is the JSON output that is generated from
 * /WEB-INF/templates/administrator/default/json/apiResponse.jsp and the other files it calls.
 * 
 * @author David Tobey
 */
public final class APIFilter extends BaseFilter implements Filter {
	static Log log = LogFactory.getLog(APIFilter.class);

	private FilterConfig filterConfig = null;

	public void init(FilterConfig filterConfig) throws ServletException {
		this.filterConfig = filterConfig;
	}

	public void destroy() {
	}

	public void doFilter(ServletRequest request, ServletResponse response, FilterChain chain) throws IOException,
			ServletException {

		// Transalate API urls. Pass the request along right away if it ends in .do, or is an image.
		HttpServletRequest req = (HttpServletRequest) request;
		if (req.getServletPath().endsWith(".do") || req.getServletPath().endsWith(".jsp")
				|| req.getServletPath().endsWith(".html") || isStaticRequest(req)) {
			chain.doFilter(request, response);
			return;
		}
		if (req.getServletPath().startsWith("/api/")) {

			// Try grabbing the settings from the application scope; forward to Welcome.do if we can't.
			Settings settings = (Settings) filterConfig.getServletContext().getAttribute("settings");
			if (settings == null) {
				log.info("Can't perform API path translation for " + req.getServletPath() + req.getQueryString()
						+ " because the application's settings have not been initialized. Returning a 503 response.");
				((HttpServletResponse) response).setStatus(HttpServletResponse.SC_SERVICE_UNAVAILABLE);
				return;
			} else if (settings.getValue("enableAPIRequests") == null
					|| settings.getValue("enableAPIRequests").equals("0")) {
				log.info("Global setting for API requests is off. " + req.getServletPath() + req.getQueryString()
						+ ". Returning a 403 response.");
				((HttpServletResponse) response).setStatus(HttpServletResponse.SC_FORBIDDEN);
				return;
			}

			// Set a flag that this is an API request. This signals AdministratorRequestProcessor to authenticate the
			// request with API credentials, and it signals BaseRequestProcessor to forward the response to
			// apiResponses.jsp.
			request.setAttribute("isApiRequest", Boolean.TRUE);

			// Translate the URL. Eg. /administrator/api/product => /administrator/ProductEditForm.do
			String forwardPath = translateAPIURL(req);

			// Modify the request: If it's a PUT request, make it a POST so the Struts servlet responds to it.
			// If the body of the request is JSON, convert it into regular parameters since that's what the Struts
			// servlet requires.
			req = new APIRequestWrapper(req);
			if (forwardPath != null) {

				// Append the query string, if there was one.
				if (req.getQueryString() != null) {
					if (forwardPath.indexOf("?") != -1) {
						forwardPath = forwardPath + "&" + req.getQueryString();
					} else {
						forwardPath = forwardPath + "?" + req.getQueryString();
					}
				}

				// Save the HTTP refer so we can log it in BaseRequestProcessor
				if (req.getHeader("Referer") != null) {
					req.setAttribute("Referer", req.getHeader("Referer"));
				}

				// Forward the request.
				log.info(req.getServletPath() + " translated to: " + forwardPath);

				// Many API requests are for administrator actions. These need to be wrapped so roles can be applied.
				if (forwardPath.startsWith("/administrator")) {
					req = new AdministratorRequestWrapper(req);
					req.setAttribute("administratorRolesAdapter", new RolesPermissionsAdapter(req));
				}

				RequestDispatcher rd = req.getRequestDispatcher(forwardPath);
				rd.forward(req, response);
				return;

			}
		}
		chain.doFilter(request, response);
		return;
	}

	private String translateAPIURL(HttpServletRequest req) {
		String servletPath = req.getServletPath();
		String forwardPath = null;
		String secondElementInPath = null; // Either "administrator", or an entity, or a non-admin action
		String capitalizedEntity = null;
		String thirdElementInPath = null; // Either "view", "edit", "add", or an admin action
		String[] pathElements = servletPath.split("/");
		if (pathElements.length > 2) {
			secondElementInPath = pathElements[2];
			capitalizedEntity = WordUtils.capitalize(secondElementInPath);
			if (capitalizedEntity.equals("Sku")) {
				capitalizedEntity = "SKU";
			}
		}
		if (pathElements.length > 3) {
			thirdElementInPath = pathElements[3];
		}
		if (secondElementInPath != null && secondElementInPath.equals("public") && thirdElementInPath != null) {

			// Customer-side actions. Eg /api/public/CartAdd => /CartAdd.do
			forwardPath = "/" + thirdElementInPath + ".do";

			// Tell apiResponse.jsp which action this was
			req.setAttribute("apiAction", thirdElementInPath);

		} else if (secondElementInPath != null && secondElementInPath.equals("administrator")
				&& thirdElementInPath != null) {

			// Admin-side actions. Eg /api/administrator/SomeAdminAction => /administrator/SomeAdminAction.do
			forwardPath = "/administrator/" + thirdElementInPath + ".do";

			// Tell apiResponse.jsp which admin action this was
			req.setAttribute("apiAdministratorAction", thirdElementInPath);

		} else if (secondElementInPath != null && thirdElementInPath != null && thirdElementInPath.equals("list")) {

			// Eg /api/product/list => /administrator/Product.do
			forwardPath = "/administrator/" + capitalizedEntity + ".do";

			// Tell apiResponse.jsp which admin action this was
			req.setAttribute("apiAdministratorAction", "list");

		} else if (secondElementInPath != null) {

			// Admin-side CRUD actions.
			if (req.getMethod().equals("POST")) {

				// Eg /api/product (POST) => /administrator/ProductAdd.do
				forwardPath = "/administrator/" + capitalizedEntity + "Add.do";

				// Special case where OrderAddForm adds an order and forwards to order.orderOverview
				if (secondElementInPath.equals("order")) {
					forwardPath = "/administrator/" + capitalizedEntity + "AddForm.do";
				}

				// Used by AddEdit Form classes to know if the entity is being added or edited and validate accordingly.
				req.setAttribute("addSubmit", Boolean.TRUE);
			} else if (req.getMethod().equals("GET")) {

				// Eg /api/product (GET) => /administrator/ProductEditForm.do
				forwardPath = "/administrator/" + capitalizedEntity + "EditForm.do";
			} else if (req.getMethod().equals("PUT") || req.getMethod().equals("PATCH")) {

				// Eg /api/product (PATCH) => /administrator/ProductEdit.do
				forwardPath = "/administrator/" + capitalizedEntity + "Edit.do";
			} else if (req.getMethod().equals("DELETE")) {

				// Eg /api/product (DELETE) => /administrator/ProductDelete.do
				forwardPath = "/administrator/" + capitalizedEntity + "Delete.do";
			}

			// Tell apiResponse.jsp which action this was
			req.setAttribute("apiAction", secondElementInPath);

			// Tell apiResponse.jsp which entity to render
			req.setAttribute("entity", secondElementInPath);
		} else {
			log.warn("Could not translate API path: " + servletPath);
		}
		return forwardPath;
	}
}
