package com.softslate.commerce.customer.core;

import java.io.InputStream;
import java.util.Collection;
import java.util.Map;

import javax.servlet.http.HttpServletRequest;
import javax.servlet.http.HttpServletResponse;

import org.apache.struts.action.ActionMessages;

import com.softslate.commerce.businessobjects.core.ApplicationMessage;
import com.softslate.commerce.businessobjects.customer.Customer;

/**
 * Interface for utility methods used by Struts Action classes.
 * <p>
 * When the application needs to create an instance that implements
 * <code>ActionUtils</code>, it calls
 * {@link com.softslate.commerce.businessobjects.core.BusinessObjectFactory#createUtilsObject}
 * , which finds the name of the Java class to instantiate from the
 * "actionUtilsImplemeter" setting in the <code>appComponents.properties</code>
 * file.
 * <p>
 * The default "actionUtilsImplemeter" is
 * {@link com.softslate.commerce.customer.core.ActionUtilsImpl}.
 * 
 * @author David Tobey
 * 
 */
public interface ActionUtils {

	/**
	 * Parses a string template with variables contained in a given
	 * <code>BaseForm</code>. Used to parse email templates for dynamically
	 * generated emails the store generates.
	 * <p>
	 * The string template may have a number of placeholders deliniated by "%%".
	 * (E.g. "%%STORE_NAME%%"). This method replaces the placeholders with
	 * values from the <code>BaseForm</code>, be they settings from the database
	 * such as the store's name, or values from the current user's order.
	 * <p>
	 * Note that with the release of 2.0, order invoice emails are now generated
	 * from JSP templates in the <code>/WEB-INF/layouts/default/order</code>
	 * directory, rather than by using this method.
	 * 
	 * @param baseForm
	 *            Contains variables that will be used to replace the template's
	 *            placeholders.
	 * @param template
	 *            A <code>String</code> containing placeholders that this method
	 *            will replace.
	 * @return The parsed template.
	 */
	abstract String parseGenericTemplate(BaseForm baseForm, String template);

	/**
	 * Parses a string template with variables contained in a given
	 * <code>BaseForm</code> and
	 * {@link com.softslate.commerce.businessobjects.customer.Customer}. Used to
	 * parse the lost password email template.
	 * <p>
	 * The string template may have a number of placeholders deliniated by "%%".
	 * (E.g. "%%USERNAME%%"). This method replaces the placeholders with values
	 * from the <code>BaseForm</code> and <code>Customer</code>, be they
	 * settings from the database such as the store's name, or the current
	 * customer's user name and password.
	 * 
	 * @param customer
	 *            The current <code>Customer</code> whose lost password is being
	 *            emailed.
	 * @param baseForm
	 *            Contains variables that will be used to replace the template's
	 *            placeholders.
	 * @param template
	 *            The template for the lost password email, which is being
	 *            parsed.
	 * @return The parsed template.
	 */
	abstract String parseLostPasswordTemplate(Customer customer,
			BaseForm baseForm, String template);

	/**
	 * Parses a string template with variables contained in a given
	 * <code>BaseForm</code> and a <code>Map</code> holding information about a
	 * SKU. Used to parse the low stock email template.
	 * <p>
	 * 
	 * @param baseForm
	 *            Contains variables that will be used to replace the template's
	 *            placeholders.
	 * @param lowStockEmail
	 *            A <code>Map</code> containing information about a SKU that has
	 *            hit its low stock level.
	 * @param template
	 *            The template for the low stock email, which is being parsed.
	 * @return The parsed template.
	 */
	abstract String parseLowStockTemplate(BaseForm baseForm, Map lowStockEmail,
			String template);

	/**
	 * Sends a multipart text and HTML email from the store. Used for sending
	 * order notifications, invoices, lost password emails, and low stock
	 * emails.
	 * 
	 * @param toEmails
	 *            A comma-separated list of email addresses the email is being
	 *            sent to.
	 * @param ccEmails
	 *            A comma-separated list of email addresses the email is being
	 *            copied to.
	 * @param bccEmails
	 *            A comma-separated list of email addresses the email is being
	 *            blind copied to.
	 * @param fromEmail
	 *            The email address used as the email's from address.
	 * @param smtpServer
	 *            The SMTP server to be used to send the email.
	 * @param subject
	 *            The subject used for the email.
	 * @param textMessage
	 *            The message for the text part of the email.
	 * @param htmlMessage
	 *            The message for the HTML part of the email.
	 * @param username
	 *            Username for SMTP authentication, or null if no authentication
	 *            is required.
	 * @param password
	 *            Password for SMTP authentication, or null if no authentication
	 *            is required.
	 * @throws Exception
	 */
	abstract void sendMultipartEmail(String toEmails, String ccEmails,
			String bccEmails, String fromEmail, String smtpServer,
			String subject, String textMessage, String htmlMessage,
			String username, String password) throws Exception;

	/**
	 * Sends a multipart text and HTML email from the store. Used for sending
	 * order notifications, invoices, lost password emails, and low stock
	 * emails.
	 * 
	 * @param toEmails
	 *            A comma-separated list of email addresses the email is being
	 *            sent to.
	 * @param ccEmails
	 *            A comma-separated list of email addresses the email is being
	 *            copied to.
	 * @param bccEmails
	 *            A comma-separated list of email addresses the email is being
	 *            blind copied to.
	 * @param fromEmail
	 *            The email address used as the email's from address.
	 * @param smtpServer
	 *            The SMTP server to be used to send the email.
	 * @param subject
	 *            The subject used for the email.
	 * @param textMessage
	 *            The message for the text part of the email.
	 * @param htmlMessage
	 *            The message for the HTML part of the email.
	 * @param username
	 *            Username for SMTP authentication, or null if no authentication
	 *            is required.
	 * @param password
	 *            Password for SMTP authentication, or null if no authentication
	 *            is required.
	 * @param characterEncoding
	 *            Optional character encoding to be used for the email. If null
	 *            or empty, no particular encoding is specified.
	 * @throws Exception
	 */
	abstract void sendMultipartEmail(String toEmails, String ccEmails,
			String bccEmails, String fromEmail, String smtpServer,
			String subject, String textMessage, String htmlMessage,
			String username, String password, String smtpPort,
			String characterEncoding) throws Exception;

	/**
	 * Sends a multipart text and HTML email from the store. Used for sending
	 * order notifications, invoices, lost password emails, and low stock
	 * emails.
	 * 
	 * @param toEmails
	 *            A comma-separated list of email addresses the email is being
	 *            sent to.
	 * @param ccEmails
	 *            A comma-separated list of email addresses the email is being
	 *            copied to.
	 * @param bccEmails
	 *            A comma-separated list of email addresses the email is being
	 *            blind copied to.
	 * @param fromEmail
	 *            The email address used as the email's from address.
	 * @param smtpServer
	 *            The SMTP server to be used to send the email.
	 * @param subject
	 *            The subject used for the email.
	 * @param textMessage
	 *            The message for the text part of the email.
	 * @param htmlMessage
	 *            The message for the HTML part of the email.
	 * @param username
	 *            Username for SMTP authentication, or null if no authentication
	 *            is required.
	 * @param password
	 *            Password for SMTP authentication, or null if no authentication
	 *            is required.
	 * @param characterEncoding
	 *            Optional character encoding to be used for the email. If null
	 *            or empty, no particular encoding is specified.
	 * @param attachmentName
	 *            Name of the attachment (report.xls)
	 * @param attachment
	 *            Byte array of data for simple attachments
	 * @param attachmentType
	 *            The mime type of the attachment (application/vnd.ms-excel)
	 * @throws Exception
	 */
	abstract void sendMultipartEmail(String toEmails, String ccEmails,
			String bccEmails, String fromEmail, String smtpServer,
			String subject, String textMessage, String htmlMessage,
			String username, String password, String smtpPort,
			String characterEncoding, String attachmentName, byte[] attachment,
			String attachmentType) throws Exception;

	/**
	 * Determines what checkout screen to direct the user to based on the status
	 * of the session and the store's database configuration. The possible
	 * screen names returned are:
	 * <ul>
	 * <li>"catalogMode", if the store's "catalogMode" database setting is set.</li>
	 * <li>"cart", if the user's cart is empty.</li>
	 * <li>"forceLogin", if the "forceLogin" database setting is set and the
	 * user is not logged in.</li>
	 * <li>"inviteLogin", if the "inviteLogin" database setting is set and the
	 * user has not declined to log in and is not logged in.</li>
	 * <li>"checkoutAddresses", if the user has not entered in all the required
	 * address fields for the order, and the "useComboScreen" setting is not
	 * set.</li>
	 * <li>"checkoutPayment", if payment information is required, the total of
	 * the current order is greater than zero, and the "useComboScreen" setting
	 * is not set.</li>
	 * <li>"checkoutCombo", if the user has not entered in all the required
	 * address fields, or if payment information is required, and if the
	 * "useComboScreen" setting is set.</li>
	 * <li>"checkoutConfirm", if none of the other screens apply.</li>
	 * </ul>
	 * <p>
	 * The above screen names correspond to
	 * <code>ActionForward</codes>s defined in the
	 * "/WEB-INF/conf/order/struts-config-order.xml" file.
	 * 
	 * @param baseForm
	 *            The form class corresponding to the current request. Holds
	 *            various properties of the request to help action classes
	 *            communicate with the rest of the application.
	 * @return A string identifying the name of the checkout screen to direct
	 *         the user to.
	 * @throws Exception
	 */
	abstract String findCheckoutScreen(BaseForm baseForm) throws Exception;

	/**
	 * Validates the required billing address fields in the user's cart. Loops
	 * through the "billingRequired" database setting to check which fields are
	 * required.
	 * 
	 * @param baseForm
	 *            The form class corresponding to the current request. Holds
	 *            various properties of the request to help action classes
	 *            communicate with the rest of the application.
	 * @return <code>true</code> if all the required billing fields are present
	 *         in the user's cart. <code>false</code> if not.
	 * @throws Exception
	 */
	abstract boolean validateBillingFields(BaseForm baseForm) throws Exception;

	/**
	 * Validates the required delivery address fields in the user's cart. Loops
	 * through the "deliveryRequired" database setting to check which fields are
	 * required.
	 * 
	 * @param baseForm
	 *            The form class corresponding to the current request. Holds
	 *            various properties of the request to help action classes
	 *            communicate with the rest of the application.
	 * @return <code>true</code> if all the required billing fields are present
	 *         in the user's cart. <code>false</code> if not.
	 * @throws Exception
	 */
	abstract boolean validateDeliveryFields(BaseForm baseForm) throws Exception;

	/**
	 * Eliminates non-standard characters from a file name so it can be safely
	 * saved across various file systems.
	 * 
	 * @param originalFileName
	 * @return A modified version of the file name with non-standard characters
	 *         eliminated
	 * @throws Exception
	 */
	abstract String makeFilenameNice(String originalFileName) throws Exception;

	/**
	 * Reads a text file and returns its contents as a string.
	 * 
	 * @param filename
	 * @return The file's contents as a string
	 * @throws Exception
	 */
	abstract String readFile(String filename) throws Exception;

	/**
	 * Writes a text file to a given location.
	 * 
	 * @param filename
	 * @param fileContents
	 * @return Always true if no exceptions
	 * @throws Exception
	 */
	abstract boolean writeFile(String filename, String fileContents)
			throws Exception;

	/**
	 * Saves an input stream as a file at a given location.
	 * 
	 * @param filename
	 * @param file
	 * @return true, if saving the file succeeded
	 * @throws Exception
	 */
	abstract boolean saveFile(String filename, InputStream file)
			throws Exception;

	/**
	 * Runs at the beggining of each request, after system objects have been
	 * initialized. Called from the BaseRequestProcessor.
	 * 
	 * @param request
	 * @param response
	 * @return true, if the request should proceed
	 * @throws Exception
	 */
	abstract boolean onPreRequestEnd(HttpServletRequest request,
			HttpServletResponse response);

	/**
	 * Runs at the beggining of each request to the customer interface, after
	 * system objects have been initialized. Called from the
	 * CustomerRequestProcessor.
	 * 
	 * @param request
	 * @param response
	 * @return true, if the request should proceed
	 * @throws Exception
	 */
	abstract boolean onCustomerPreRequestEnd(HttpServletRequest request,
			HttpServletResponse response);

	/**
	 * Runs at the beggining of each request to the administrator interface,
	 * after system objects have been initialized. Called from the
	 * CustomerRequestProcessor.
	 * 
	 * @param request
	 * @param response
	 * @return true, if the request should proceed
	 * @throws Exception
	 */
	abstract boolean onAdministratorPreRequestEnd(HttpServletRequest request,
			HttpServletResponse response);

	/**
	 * If the system settings permit it, add a persistent cookie to the user's
	 * browser identifying the cart so it can be loaded on a later visit.
	 * <p>
	 * A cookie with the name "sscsctoken" and the value of the user's session
	 * id will be saved in the browser and recorded in the sscOrder database
	 * table's "savedCartToken" field.
	 * 
	 * @param baseForm
	 *            The instance of <code>CartAddForm</code> corresponding to the
	 *            request.
	 * @param parameters
	 *            The parameters of the add to cart request.
	 * @param response
	 *            The servlet's response object, used to save the cookie to the
	 *            browser.
	 */
	public abstract void processSavedCartCookie(BaseForm baseForm,
			Map parameters, HttpServletResponse response);

	/**
	 * Process the results of adding an item or items to the cart. This method
	 * first sends out low stock inventory emails. It then adds error messages
	 * for customers if required attributes were missing from the request or the
	 * item or items are out of stock.
	 * <p>
	 * In the case of an error, the attribute "errorProductCode" is set in the
	 * request, corresponding to the product code of the first item being added
	 * to the cart. If the submission came from a product page, this will tell
	 * the <code>ProductForm</code> to populate the request with that product's
	 * data, so the errors can be displayed on the product page.
	 * <p>
	 * Finally, this method calls the <code>addInventoryMessages</code> method,
	 * to populate the request with (non-error) messages relating to inventory.
	 * 
	 * @param results
	 *            The Map of results returned by the call to the
	 *            <code>execute</code> method of this class.
	 * @param baseForm
	 *            The instance of <code>CartAddForm</code> corresponding to the
	 *            request.
	 * @return A String, "success" or "failure", corresponding to the
	 *         <code>ActionFoward</code> that the <code>execute</code> method
	 *         will return.
	 */
	public abstract String processCartAddResults(Map results, BaseForm baseForm);

	/**
	 * Adds non-error messages to the request related to inventory. Two types of
	 * messages may be added.
	 * <ul>
	 * <li>Messages indicating that the quantity of an item was more than its
	 * inventory level and that therefore the quantity was automatically lowered
	 * to match the amount in stock.</li>
	 * <li>Messages indicating that with this request a product has just hit its
	 * out of stock level.</li>
	 * </ul>
	 * 
	 * @param results
	 *            The Map of results returned by the call to the
	 *            <code>execute</code> method of this class.
	 * @param baseForm
	 *            The instance of <code>CartAddForm</code> corresponding to the
	 *            current request.
	 */
	public abstract void addInventoryMessages(Map results, BaseForm baseForm);

	/**
	 * Load a set of estimated shipping options for the current cart. The
	 * options may be merely estimates because the shopper may not have defined
	 * a delivery address yet, if he has not logged in or entered addresses
	 * during checkout. In these cases, the default city, state, country, and
	 * ZIP defined in the database settings are used.
	 * 
	 * @param baseForm
	 * @return A Collection of Maps representing the various shipping options
	 *         and their rates.
	 */
	public abstract Collection loadEstimatedShippingOptions(BaseForm baseForm);

	/**
	 * Creates a string-based Map based on all the request parameters.
	 * 
	 * @param request
	 * @return A map representing all the request parameters.
	 * @throws Exception
	 */
	public abstract Map buildParameterMap(HttpServletRequest request)
			throws Exception;

	/**
	 * Create log entries in the application's audit logs, containing all the
	 * key information about each request.
	 * 
	 * @param request
	 */
	public abstract void processAuditLogging(HttpServletRequest request);

	/**
	 * Given a request object, converts all the incoming parameters to a string
	 * for debugging. The values of certain parameters such as passwords are
	 * masked for security.
	 * 
	 * @param request
	 *            The current request object.
	 * @return A String representing the name value pairs of the request
	 *         parameters.
	 */
	public abstract String getRequestParamsDebugMessage(
			HttpServletRequest request);

	/**
	 * Returns the URL link to look up a given tracking number's status, on
	 * either the UPS, USPS, or FedEx website.
	 * 
	 * @param trackingNumber
	 * @return The URL for the tracking number on the UPS, USPS, or FedEx
	 *         website
	 */
	public abstract String createTrackingNumberLink(String trackingNumber);

	/**
	 * Given a String name, creates a URL-safe code that can be used for the
	 * "code" field of many domain entities that require it.
	 */
	public abstract String generateCodeFromName(String name);

	abstract void processApplicationMessages(ActionMessages errors, ActionMessages messages,
			Collection<ApplicationMessage> applicationMessages);

	abstract String parseLostPasswordTemplateAdvanced(Customer customer, BaseForm baseForm, String template,
			String token);

}
